// Package addrtype provides Address type assigned numbers.
package addrtype

// Address type IDs.
const (
	IPv4          int32 = 2
	Directional   int32 = 3
	ChaosNet      int32 = 5
	XNS           int32 = 6
	ISO           int32 = 7
	DECNETPhaseIV int32 = 12
	AppleTalkDDP  int32 = 16
	NetBios       int32 = 20
	IPv6          int32 = 24
)
